@@ -10,7 +10,9 @@ use resolve::PluginCache;
 use swc_common::{
     collections::AHashMap,
     errors::{Diagnostic, HANDLER},
+    hygiene::MutableMarkContext,
     plugin::{PluginError, Serialized},
+    Mark, SyntaxContext,
 };
 use wasmer::{
     imports, Array, Exports, Function, Instance, LazyInit, Memory, Module, Store, WasmPtr,
@@ -36,6 +38,43 @@ fn copy_bytes_into_host(memory: &Memory, bytes_ptr: i32, bytes_ptr_len: i32) ->
         .collect::<Vec<u8>>()
 }
 
+/// Locate a view from given memory, write serialized bytes into.
+fn write_into_memory_view<F>(
+    memory: &Memory,
+    serialized_bytes: &Serialized,
+    get_allocated_ptr: F,
+) -> (i32, i32)
+where
+    F: Fn(usize) -> i32,
+{
+    let serialized = serialized_bytes.as_ref();
+    let serialized_len = serialized.len();
+
+    let ptr_start = get_allocated_ptr(serialized_len);
+    let ptr_start_size = ptr_start
+        .try_into()
+        .expect("Should be able to convert to usize");
+
+    // Note: it's important to get a view from memory _after_ alloc completes
+    let view = memory.view::<u8>();
+
+    // loop over the Wasm memory view's bytes, assign bytes value of alignedvec from
+    // serialized
+    for (cell, byte) in view[ptr_start_size..ptr_start_size + serialized_len + 1]
+        .iter()
+        .zip(serialized.iter())
+    {
+        cell.set(*byte)
+    }
+
+    (
+        ptr_start,
+        serialized_len
+            .try_into()
+            .expect("Should be able to convert to i32"),
+    )
+}
+
 /// Set plugin's transformed result into host's enviroment.
 /// This is an `imported` fn - when we instantiate plugin module, we inject this
 /// fn into pluging's export space. Once transform completes, plugin will call
@@ -63,6 +102,95 @@ fn emit_diagnostics(env: &HostEnvironment, bytes_ptr: i32, bytes_ptr_len: i32) {
     }
 }
 
+/// A proxy to Mark::fresh() that can be used in plugin.
+/// This it not direcly called by plugin, instead `impl Mark` will selectively
+/// call this depends on the running context.
+fn mark_fresh_proxy(parent: u32) -> u32 {
+    Mark::fresh(Mark::from_u32(parent)).as_u32()
+}
+
+fn mark_parent_proxy(self_mark: u32) -> u32 {
+    Mark::from_u32(self_mark).parent().as_u32()
+}
+
+fn mark_is_builtin_proxy(self_mark: u32) -> u32 {
+    Mark::from_u32(self_mark).is_builtin() as u32
+}
+
+fn mark_set_builtin_proxy(self_mark: u32, is_builtin: u32) {
+    Mark::from_u32(self_mark).set_is_builtin(is_builtin != 0);
+}
+
+/// A proxy to Mark::is_descendant_of_() that can be used in plugin.
+/// Origianl call site have mutable param, which we'll pass over as return value
+/// via serialized MutableMarkContext.
+/// Inside of guest context, once this host function returns it'll assign params
+/// with return value accordingly.
+fn mark_is_descendant_of_proxy(
+    env: &HostEnvironment,
+    self_mark: u32,
+    ancestor: u32,
+    allocated_ptr: i32,
+) {
+    let self_mark = Mark::from_u32(self_mark);
+    let ancestor = Mark::from_u32(ancestor);
+
+    let return_value = self_mark.is_descendant_of(ancestor);
+
+    if let Some(memory) = env.memory_ref() {
+        let serialized_bytes = Serialized::serialize(&MutableMarkContext(
+            self_mark.as_u32(),
+            0,
+            return_value as u32,
+        ))
+        .expect("Should be serializable");
+
+        write_into_memory_view(memory, &serialized_bytes, |_| allocated_ptr);
+    }
+}
+
+fn mark_least_ancestor_proxy(env: &HostEnvironment, a: u32, b: u32, allocated_ptr: i32) {
+    let a = Mark::from_u32(a);
+    let b = Mark::from_u32(b);
+
+    let return_value = Mark::least_ancestor(a, b).as_u32();
+
+    if let Some(memory) = env.memory_ref() {
+        let serialized_bytes =
+            Serialized::serialize(&MutableMarkContext(a.as_u32(), b.as_u32(), return_value))
+                .expect("Should be serializable");
+
+        write_into_memory_view(memory, &serialized_bytes, |_| allocated_ptr);
+    }
+}
+
+fn syntax_context_apply_mark_proxy(self_syntax_context: u32, mark: u32) -> u32 {
+    SyntaxContext::from_u32(self_syntax_context)
+        .apply_mark(Mark::from_u32(mark))
+        .as_u32()
+}
+
+fn syntax_context_remove_mark_proxy(env: &HostEnvironment, self_mark: u32, allocated_ptr: i32) {
+    let mut self_mark = SyntaxContext::from_u32(self_mark);
+
+    let return_value = self_mark.remove_mark();
+
+    if let Some(memory) = env.memory_ref() {
+        let serialized_bytes = Serialized::serialize(&MutableMarkContext(
+            self_mark.as_u32(),
+            0,
+            return_value.as_u32(),
+        ))
+        .expect("Should be serializable");
+
+        write_into_memory_view(memory, &serialized_bytes, |_| allocated_ptr);
+    }
+}
+
+fn syntax_context_outer_proxy(self_mark: u32) -> u32 {
+    SyntaxContext::from_u32(self_mark).outer().as_u32()
+}
+
 #[derive(wasmer::WasmerEnv, Clone)]
 /// An external enviornment state imported (declared in host, injected into
 /// guest) fn can access. This'll allow host access updated state via plugin's
@@ -178,6 +306,43 @@ fn load_plugin(
                 emit_diagnostics,
             );
 
+            let mark_fresh_fn_decl = Function::new_native(&wasmer_store, mark_fresh_proxy);
+            let mark_parent_fn_decl = Function::new_native(&wasmer_store, mark_parent_proxy);
+            let mark_is_builtin_fn_decl =
+                Function::new_native(&wasmer_store, mark_is_builtin_proxy);
+            let mark_set_builtin_fn_decl =
+                Function::new_native(&wasmer_store, mark_set_builtin_proxy);
+            let mark_is_descendant_of_fn_decl = Function::new_native_with_env(
+                &wasmer_store,
+                HostEnvironment {
+                    memory: LazyInit::default(),
+                    transform_result: transform_result.clone(),
+                },
+                mark_is_descendant_of_proxy,
+            );
+
+            let mark_least_ancestor_fn_decl = Function::new_native_with_env(
+                &wasmer_store,
+                HostEnvironment {
+                    memory: LazyInit::default(),
+                    transform_result: transform_result.clone(),
+                },
+                mark_least_ancestor_proxy,
+            );
+
+            let syntax_context_apply_mark_fn_decl =
+                Function::new_native(&wasmer_store, syntax_context_apply_mark_proxy);
+            let syntax_context_remove_mark_fn_decl = Function::new_native_with_env(
+                &wasmer_store,
+                HostEnvironment {
+                    memory: LazyInit::default(),
+                    transform_result: transform_result.clone(),
+                },
+                syntax_context_remove_mark_proxy,
+            );
+            let syntax_context_outer_fn_decl =
+                Function::new_native(&wasmer_store, syntax_context_outer_proxy);
+
             // Plugin binary can be either wasm32-wasi or wasm32-unknown-unknown
             let import_object = if is_wasi_module(&module) {
                 // Create the `WasiEnv`.
@@ -197,6 +362,26 @@ fn load_plugin(
                 let mut env = Exports::new();
                 env.insert("__set_transform_result", set_transform_result_fn_decl);
                 env.insert("__emit_diagnostics", emit_diagnostics_fn_decl);
+
+                env.insert("__mark_fresh_proxy", mark_fresh_fn_decl);
+                env.insert("__mark_parent_proxy", mark_parent_fn_decl);
+                env.insert("__mark_is_builtin_proxy", mark_is_builtin_fn_decl);
+                env.insert("__mark_set_builtin_proxy", mark_set_builtin_fn_decl);
+                env.insert(
+                    "__mark_is_descendant_of_proxy",
+                    mark_is_descendant_of_fn_decl,
+                );
+                env.insert("__mark_least_ancestor", mark_least_ancestor_fn_decl);
+                env.insert(
+                    "__syntax_context_apply_mark_proxy",
+                    syntax_context_apply_mark_fn_decl,
+                );
+                env.insert(
+                    "__syntax_context_remove_mark_proxy",
+                    syntax_context_remove_mark_fn_decl,
+                );
+                env.insert("__syntax_context_outer_proxy", syntax_context_outer_fn_decl);
+
                 import_object.register("env", env);
                 import_object
             }
@@ -206,7 +391,16 @@ fn load_plugin(
                 imports! {
                     "env" => {
                         "__set_transform_result" => set_transform_result_fn_decl,
-                        "__emit_diagnostics" => emit_diagnostics_fn_decl
+                        "__emit_diagnostics" => emit_diagnostics_fn_decl,
+                        "__mark_fresh_proxy" => mark_fresh_fn_decl,
+                        "__mark_parent_proxy" => mark_parent_fn_decl,
+                        "__mark_is_builtin_proxy" => mark_is_builtin_fn_decl,
+                        "__mark_set_builtin_proxy" => mark_set_builtin_fn_decl,
+                        "__mark_is_descendant_of_proxy" => mark_is_descendant_of_fn_decl,
+                        "__mark_least_ancestor" => mark_least_ancestor_fn_decl,
+                        "__syntax_context_apply_mark_proxy" => syntax_context_apply_mark_fn_decl,
+                        "__syntax_context_remove_mark_proxy" => syntax_context_remove_mark_fn_decl,
+                        "__syntax_context_outer_proxy" => syntax_context_outer_fn_decl
                     }
                 }
             };
@@ -266,27 +460,11 @@ impl PluginTransformTracker {
     ) -> Result<(i32, i32), Error> {
         let memory = self.instance.exports.get_memory("memory")?;
 
-        let serialized = serialized_bytes.as_ref();
-        let serialized_len = serialized.len();
-
-        let allocated_ptr = self
-            .exported_plugin_alloc
-            .call(serialized_len.try_into()?)?;
-
-        // Note: it's important to get a view from memory _after_ alloc completes
-        let view = memory.view::<u8>();
-
-        // loop over the Wasm memory view's bytes, assign bytes value of alignedvec from
-        // serialized
-        let ptr_start: usize = allocated_ptr.try_into()?;
-        for (cell, byte) in view[ptr_start..ptr_start + serialized_len + 1]
-            .iter()
-            .zip(serialized.iter())
-        {
-            cell.set(*byte)
-        }
-
-        let ptr = (allocated_ptr, serialized_len.try_into()?);
+        let ptr = write_into_memory_view(memory, serialized_bytes, |serialized_len| {
+            self.exported_plugin_alloc
+                .call(serialized_len.try_into().expect(""))
+                .expect("")
+        });
 
         self.allocated_ptr_vec.push(ptr);
         Ok(ptr)
